# -*- encoding: utf-8
"""
This file maps ugly curses keycodes to human readable versions.
"""

import curses


key_map = {
    # Single keys
    curses.KEY_F1: "f1",    # 265
    curses.KEY_F2: "f2",    # 266
    curses.KEY_F3: "f3",    # 267
    curses.KEY_F4: "f4",    # 268
    curses.KEY_F5: "f5",    # 269
    curses.KEY_F6: "f6",    # 270
    curses.KEY_F7: "f7",    # 271
    curses.KEY_F8: "f8",    # 272
    curses.KEY_F9: "f9",    # 273
    curses.KEY_F10: "f10",  # 274
    curses.KEY_F11: "f11",  # 275
    curses.KEY_F12: "f12",  # 276
    "KEY_F(1)": "f1",
    "KEY_F(2)": "f2",
    "KEY_F(3)": "f3",
    "KEY_F(4)": "f4",
    "KEY_F(5)": "f5",
    "KEY_F(6)": "f6",
    "KEY_F(7)": "f7",
    "KEY_F(8)": "f8",
    "KEY_F(9)": "f9",
    "KEY_F(10)": "f10",
    "KEY_F(11)": "f11",
    "KEY_F(12)": "f12",


    curses.KEY_UP: "up",
    curses.KEY_DOWN: "down",
    curses.KEY_LEFT: "left",
    curses.KEY_RIGHT: "right",
    "KEY_UP": "up",
    "KEY_DOWN": "down",
    "KEY_LEFT": "left",
    "KEY_RIGHT": "right",

    curses.KEY_ENTER: "enter",
    "KEY_ENTER": "enter",
    "\n": "enter",
    "^J": "enter",
    "^M": "enter",
    343: "shift+enter",

    curses.KEY_BACKSPACE: "backspace",
    "KEY_BACKSPACE": "backspace",
    "^?": "backspace",
    "^H": "backspace", # on Windows

    curses.KEY_DC: "delete",
    curses.KEY_HOME: "home",
    curses.KEY_END: "end",
    curses.KEY_PPAGE: "pageup",
    curses.KEY_NPAGE: "pagedown",
    "KEY_DC": "delete",
    "KEY_HOME": "home",
    "KEY_END": "end",
    "KEY_PPAGE": "pageup",
    "KEY_NPAGE": "pagedown",

    curses.KEY_IC: "insert",
    "KEY_IC": "insert",
    331: "insert",
    "\t": "tab",
    "^I": "tab",
    "^[": "escape",


    # Control
    "^A": "ctrl+a",
    "^B": "ctrl+b",
    "^C": "ctrl+c",
    "^D": "ctrl+d",
    "^E": "ctrl+e",
    "^F": "ctrl+f",
    "^G": "ctrl+g",
    # "^H": "ctrl+h",  # Conflicts with 'backspace' on Windows
    # "^I": "ctrl+i",  # Conflicts with 'tab'
    # "^J": "ctrl+j",  # Conflicts with 'enter'
    "^K": "ctrl+k",
    "^L": "ctrl+l",
    # "^M": "ctrl+m",  # Conflicts with 'enter'
    "^N": "ctrl+n",
    "^O": "ctrl+o",
    "^P": "ctrl+p",
    "^Q": "ctrl+q",
    "^R": "ctrl+r",
    "^S": "ctrl+s",
    "^T": "ctrl+t",
    "^U": "ctrl+u",
    "^V": "ctrl+v",
    "^W": "ctrl+w",
    "^X": "ctrl+x",
    "^Y": "ctrl+y",
    "^Z": "ctrl+z",  # Conflicts with suspend

    544: "ctrl+left",
    559: "ctrl+right",
    565: "ctrl+up",
    524: "ctrl+down",
    "kLFT5": "ctrl+left",
    "kRIT5": "ctrl+right",
    "kUP5": "ctrl+up",
    "kDN5": "ctrl+down",

    "kIC5": "ctrl+insert",
    "kDC5": "ctrl+delete",
    "kHOM5": "ctrl+home",
    "kEND5": "ctrl+end",

    554: "ctrl+pageup",
    549: "ctrl+pagedown",
    "kNXT5": "ctrl+pageup",
    "kPRV5": "ctrl+pagedown",

    "O5P": "ctrl+f1",
    "O5Q": "ctrl+f2",
    "O5R": "ctrl+f3",
    "O5S": "ctrl+f4",

    "KEY_F(29)": "ctrl+f5",
    "KEY_F(30)": "ctrl+f6",
    "KEY_F(31)": "ctrl+f7",
    "KEY_F(32)": "ctrl+f8",
    "KEY_F(33)": "ctrl+f9",
    "KEY_F(34)": "ctrl+f10",
    "KEY_F(35)": "ctrl+f11",
    "KEY_F(36)": "ctrl+f12",

    # Alt
    563: "alt+up",
    522: "alt+down",
    542: "alt+left",
    557: "alt+right",
    "kUP3": "alt+up",
    "kDN3": "alt+down",
    "kLFT3": "alt+left",
    "kRIT3": "alt+right",

    "kIC3": "alt+insert",
    "kDC3": "alt+delete",
    "kHOM3": "alt+home",
    "kEND3": "alt+end",

    552: "alt+pageup",
    547: "alt+pagedown",
    "kPRV3": "alt+pageup",
    "kNXT3": "alt+pagedown",

    "KEY_F(53)": "alt+f5",
    "KEY_F(54)": "alt+f6",
    "KEY_F(55)": "alt+f7",
    "KEY_F(56)": "alt+f8",
    "KEY_F(57)": "alt+f9",
    "KEY_F(58)": "alt+f10",
    "KEY_F(59)": "alt+f11",
    "KEY_F(60)": "alt+f12",

    # Shift
    curses.KEY_SLEFT: "shift+left",
    curses.KEY_SRIGHT: "shift+right",
    "KEY_SLEFT": "shift+left",
    "KEY_SRIGHT": "shift+right",
    "KEY_SR": "shift+up",
    "KEY_SF": "shift+down",

    353: "shift+tab",
    "KEY_BTAB": "shift+tab",

    "KEY_SDC": "shift+delete",
    "KEY_SHOME": "shift+home",
    "KEY_SEND": "shift+end",

    "KEY_SPREVIOUS": "shift+pageup",
    "KEY_SNEXT": "shift+pagedown",

    "O2P": "shift+f1",
    "O2Q": "shift+f2",
    "O2R": "shift+f3",
    "O2S": "shift+f4",

    "KEY_F(17)": "shift+f5",
    "KEY_F(18)": "shift+f6",
    "KEY_F(19)": "shift+f7",
    "KEY_F(20)": "shift+f8",
    "KEY_F(21)": "shift+f9",
    "KEY_F(22)": "shift+f10",
    "KEY_F(23)": "shift+f11",
    "KEY_F(24)": "shift+f12",


    # Alt Gr
    "O1P": "altgr+f1",
    "O1Q": "altgr+f2",
    "O1R": "altgr+f3",
    "O1S": "altgr+f4",


    # Shift + Alt
    "kUP4": "shift+alt+up",
    "kDN4": "shift+alt+down",
    "kLFT4": "shift+alt+left",
    "kRIT4": "shift+alt+right",

    "kIC4": "shift+alt+inset",
    "kDC4": "shift+alt+delete",
    "kHOM4": "shift+alt+home",
    "kEND4": "shift+alt+end",


    # Control + Shift
    "kUP6": "ctrl+shift+up",
    "kDN6": "ctrl+shift+down",
    "kLFT6": "ctrl+shift+left",
    "kRIT6": "ctrl+shift+right",

    "kDC6": "ctrl+shift+delete",
    "kHOM6": "ctrl+shift+home",
    "kEND6": "ctrl+shift+end",


    # Control + Alt
    "kUP7": "ctrl+alt+up",
    "kDN7": "ctrl+alt+down",
    "kLFT7": "ctrl+alt+left",
    "kRIT7": "ctrl+alt+right",

    "kPRV7": "ctrl+alt+pageup",
    "kNXT7": "ctrl+alt+pagedown",

    "kIC7": "ctrl+alt+insert",
    "kDC7": "ctrl+alt+delete",
    "kHOM7": "ctrl+alt+home",
    "kEND7": "ctrl+alt+end",

    # Special events
    "KEY_RESIZE": "resize",
}



# -*- encoding: utf-8
"""
Config handler.
"""

import os
import json
import logging

from . import suplemon_module


class Config:
    def __init__(self, app):
        self.app = app
        self.logger = logging.getLogger(__name__)
        self.default_config_filename = "defaults.json"
        self.default_keymap_filename = "keymap.json"
        self.config_filename = "suplemon-config.json"
        self.keymap_filename = "suplemon-keymap.json"
        self.home_dir = os.path.expanduser("~")
        self.config_dir = os.path.join(self.home_dir, ".config", "suplemon")

        self.defaults = {}
        self.keymap = {}
        self.config = {}
        self.key_bindings = {}

    def init(self):
        self.create_config_dir()
        return self.load_defaults()

    def path(self):
        return os.path.join(self.config_dir, self.config_filename)

    def keymap_path(self):
        return os.path.join(self.config_dir, self.keymap_filename)

    def set_path(self, path):
        parts = os.path.split(path)
        self.config_dir = parts[0]
        self.config_filename = parts[1]

    def load(self):
        path = self.path()
        config = False
        if not os.path.exists(path):
            self.logger.debug("Configuration file '{0}' doesn't exist.".format(path))
        else:
            config = self.load_config_file(path)
        if config is not False:
            self.logger.debug("Loaded configuration file '{0}'".format(path))
            self.config = self.merge_defaults(config)
        else:
            self.logger.info("Failed to load config file '{0}'.".format(path))
            self.config = dict(self.defaults)
        self.load_keys()
        return config

    def load_keys(self):
        path = self.keymap_path()
        keymap = []

        if not os.path.exists(path):
            self.logger.debug("Keymap file '{0}' doesn't exist.".format(path))
        else:
            keymap = self.load_config_file(path) or []
            if not keymap:
                self.logger.warning("Failed to load keymap file '{0}'.".format(path))

        # Build the key bindings
        # User keymap overwrites the defaults in the bindings
        self.keymap = self.normalize_keys(self.keymap + keymap)
        self.key_bindings = {}
        for binding in self.keymap:
            for key in binding["keys"]:
                self.key_bindings[key] = binding["command"]

        return True

    def normalize_keys(self, keymap):
        """Normalize the order of modifier keys in keymap."""
        modifiers = ["shift", "ctrl", "alt", "meta"]  # The modifiers in correct order
        for item in keymap:
            new_keys = []
            for key_item in item["keys"]:
                parts = key_item.split("+")
                key = parts[-1]
                if len(parts) < 2:
                    new_keys.append(key)
                    continue
                normalized = ""
                for mod in modifiers:  # Add the used modifiers back in correct order
                    if mod in parts:
                        normalized += mod + "+"
                normalized += key
                new_keys.append(normalized)
            item["keys"] = new_keys
        return keymap

    def load_defaults(self):
        if not self.load_default_config() or not self.load_default_keys():
            return False
        return True

    def load_default_config(self):
        path = os.path.join(self.app.path, "config", self.default_config_filename)
        config = self.load_config_file(path)
        if not config:
            self.logger.error("Failed to load default config file '{0}'!".format(path))
            return False
        self.defaults = config
        return True

    def load_module_configs(self):
        module_config = {}
        modules = self.app.modules.modules
        for module_name in modules.keys():
            module = modules[module_name]
            conf = module.get_default_config()
            module_config[module_name] = conf
            self.logger.debug("Loading default config for module '%s': %s" % (module_name, str(conf)))
        self.defaults["modules"] = module_config
        self.config = self.merge_defaults(self.config)

    def load_default_keys(self):
        path = os.path.join(self.app.path, "config", self.default_keymap_filename)
        config = self.load_config_file(path)
        if not config:
            self.logger.error("Failed to load default keymap file '{0}'!".format(path))
            return False
        self.keymap = config
        return True

    def reload(self):
        """Reload the config file."""
        return self.load()

    def store(self):
        """Write current config state to file."""
        data = json.dumps(self.config)
        f = open(self.config_filename)
        f.write(data)
        f.close()

    def merge_defaults(self, config):
        """Fill any missing config options with defaults."""
        return self._merge_defaults(self.defaults, config)

    def _merge_defaults(self, defaults, config):
        """Recursivley merge two dicts."""
        for key in defaults.keys():
            item = defaults[key]
            if key not in config.keys():
                config[key] = item
                continue
            if not isinstance(item, dict):
                continue
            config[key] = self._merge_defaults(item, config[key])
        return config

    def load_config_file(self, path):
        try:
            f = open(path)
            data = f.read()
            f.close()
            data = self.remove_config_comments(data)
            config = json.loads(data)
            return config
        except:
            return False

    def remove_config_comments(self, data):
        """Remove comments from a 'pseudo' JSON config file.

        Removes all lines that begin with '#' or '//' ignoring whitespace.

        :param data: Commented JSON data to clean.
        :return: Cleaned pure JSON.
        """
        lines = data.split("\n")
        cleaned = []
        for line in lines:
            line = line.strip()
            if line.startswith(("//", "#")):
                continue
            cleaned.append(line)
        return "\n".join(cleaned)

    def create_config_dir(self):
        if not os.path.exists(self.config_dir):
            try:
                os.makedirs(self.config_dir)
            except:
                self.app.logger.warning("Config folder '{0}' doesn't exist and couldn't be created.".format(
                                        self.config_dir))

    def __getitem__(self, i):
        """Get a config variable."""
        return self.config[i]

    def __setitem__(self, i, v):
        """Set a config variable."""
        self.config[i] = v

    def __str__(self):
        """Convert entire config array to string."""
        return str(self.config)

    def __len__(self):
        """Return length of top level config variables."""
        return len(self.config)


class ConfigModule(suplemon_module.Module):
    """Helper for shortcut for opening config files."""
    def init(self):
        self.config_name = "defaults.json"
        self.config_default_path = os.path.join(self.app.path, "config", self.config_name)
        self.config_user_path = self.app.config.path()

    def run(self, app, editor, args):
        if args == "defaults":
            # Open the default config in a new file only for viewing
            self.open(app, self.config_default_path, read_only=True)
        else:
            self.open(app, self.config_user_path)

    def open(self, app, path, read_only=False):
        if read_only:
            f = open(path)
            data = f.read()
            f.close()
            file = app.new_file()
            file.set_name(self.config_name)
            file.set_data(data)
            app.switch_to_file(app.last_file_index())
        else:
            # Open the user config file for editing
            f = app.file_is_open(path)
            if f:
                app.switch_to_file(app.get_file_index(f))
            else:
                if not app.open_file(path):
                    app.new_file(path)
                app.switch_to_file(app.last_file_index())



"""
Basic logging to delay printing until curses is unloaded.
"""
from __future__ import print_function
import os
import logging
from logging.handlers import BufferingHandler, RotatingFileHandler
import sys


# Define an mix of BufferingHandler and MemoryHandler which store records internally and flush on `close`
# https://docs.python.org/3.3/library/logging.handlers.html#logging.handlers.BufferingHandler
# https://docs.python.org/3.3/library/logging.handlers.html#logging.handlers.MemoryHandler
class BufferingTargetHandler(BufferingHandler):
    # Set up capacity and target for MemoryHandler
    def __init__(self, capacity, fd_target):
        """
        :param int capacity: Amount of records to store in memory
            https://github.com/python/cpython/blob/3.3/Lib/logging/handlers.py#L1161-L1176
        :param object fd_target: File descriptor to write output to (e.g. `sys.stdout`)
        """
        # Call our BufferingHandler init
        if issubclass(BufferingTargetHandler, object):
            super(BufferingTargetHandler, self).__init__(capacity)
        else:
            BufferingHandler.__init__(self, capacity)

        # Save target for later
        self._fd_target = fd_target

    def close(self):
        """Upon `close`, flush our internal info to the target"""
        # Flush our buffers to the target
        # https://github.com/python/cpython/blob/3.3/Lib/logging/handlers.py#L1185
        # https://github.com/python/cpython/blob/3.3/Lib/logging/handlers.py#L1241-L1256
        self.acquire()
        try:
            for record in self.buffer:
                if record.levelno < self.level:
                    continue
                msg = self.format(record)
                print(msg, file=self._fd_target)
        finally:
            self.release()

        # Then, run our normal close actions
        if issubclass(BufferingTargetHandler, object):
            super(BufferingTargetHandler, self).close()
        else:
            BufferingHandler.close(self)


# Initialize logging
logging.basicConfig(level=logging.NOTSET)
logger = logging.getLogger()
logger.handlers = []

# Generate and configure handlers
log_filepath = os.path.join(os.path.expanduser("~"), ".config", "suplemon", "output.log")
logger_handlers = [
    # Buffer 64k records in memory at a time
    BufferingTargetHandler(64 * 1024, fd_target=sys.stderr),
]

# Error recovery when log_filepath isn't writable
try:
    if not os.path.exists(os.path.dirname(log_filepath)):
        os.makedirs(os.path.dirname(log_filepath))
    # Output up to 4MB of records to `~/.config/suplemon/output.log` for live debugging
    # https://docs.python.org/3.3/library/logging.handlers.html#logging.handlers.RotatingFileHandler
    # DEV: We use append mode to prevent erasing out logs
    # DEV: We use 1 backup count since it won't truncate otherwise =/
    rfh = RotatingFileHandler(log_filepath, mode="a+", maxBytes=(4 * 1024 * 1024), backupCount=1)
    logger_handlers.append(rfh)
except:
    # Can't recover and can't log this error
    pass

fmt = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
logger_formatter = logging.Formatter(fmt)
for logger_handler in logger_handlers:
    logger_handler.setFormatter(logger_formatter)

# Save handlers for our logger
for logger_handler in logger_handlers:
    logger.addHandler(logger_handler)



# -*- encoding: utf-8

import difflib

from suplemon.suplemon_module import Module


class Diff(Module):
    """View a diff of the current file compared to it's on disk version."""

    def run(self, app, editor, args):
        curr_file = app.get_file()
        curr_path = curr_file.get_path()
        if not curr_path:
            self.app.set_status("File hasn't been saved, can't show diff.")
            return False

        current_data = editor.get_data()
        f = open(curr_path)
        original_data = f.read()
        f.close()
        diff = self.get_diff(original_data, current_data)

        if not diff:
            self.app.set_status("The file in the editor and on disk are identical.")
            return False
        file = app.new_file()
        file.set_name(curr_file.get_name() + ".diff")
        file.set_data(diff)
        app.switch_to_file(app.last_file_index())

    def get_diff(self, a, b):
        a = a.splitlines(1)
        b = b.splitlines(1)
        diff = difflib.unified_diff(a, b)
        return "".join(diff)


module = {
    "class": Diff,
    "name": "diff",
}



from suplemon.linelight.color_map import color_map


class Syntax:
    def get_comment(self):
        return ("-- ", "")

    def get_color(self, raw_line):
        color = color_map["white"]
        return color



# -*- encoding: utf-8
"""
Addon module loader.
"""
import os
import imp
import logging


class ModuleLoader:
    def __init__(self, app=None):
        self.app = app
        self.logger = logging.getLogger(__name__)
        # The app root directory
        self.curr_path = os.path.dirname(os.path.realpath(__file__))
        # The modules subdirectory
        self.module_path = os.path.join(self.curr_path, "modules" + os.sep)
        # Module instances
        self.modules = {}

    def load(self):
        """Find and load available modules."""
        self.logger.debug("Loading modules...")
        names = self.get_module_names()
        for name in names:
            module = self.load_single(name)
            if module:
                # Load and store the module instance
                inst = self.load_instance(module)
                if inst:
                    self.modules[module[0]] = inst

    def get_module_names(self):
        """Get names of loadable modules."""
        names = []
        dirlist = os.listdir(self.module_path)
        for item in dirlist:
            # Skip 'hidden' dot files and files beginning with and underscore
            if item.startswith((".", "_")):
                continue
            parts = item.split(".")
            if len(parts) < 2:
                # Can't find file extension
                continue
            name = parts[0]
            ext = parts[-1]
            # only load .py modules
            if ext != "py":
                continue
            names.append(name)
        return names

    def load_instance(self, module):
        """Initialize a module."""
        try:
            inst = module[1]["class"](self.app, module[0], module[1])  # Store the module instance
            return inst
        except:
            self.logger.error("Initializing module failed: {0}".format(module[0]), exc_info=True)
        return False

    def load_single(self, name):
        """Load single module file."""
        path = os.path.join(self.module_path, name+".py")
        try:
            mod = imp.load_source(name, path)
        except:
            self.logger.error("Failed loading module: {0}".format(name), exc_info=True)
            return False
        if "module" not in dir(mod):
            return False
        if "status" not in mod.module.keys():
            mod.module["status"] = False
        return name, mod.module

    def extract_docs(self):
        """Get names and docs of runnable modules and print as markdown."""
        names = sorted(self.get_module_names())
        for name in names:
            name, module = self.load_single(name)
            # Skip modules that can't be run expicitly
            if module["class"].run.__module__ == "suplemon.suplemon_module":
                continue
            # Skip undocumented modules
            if not module["class"].__doc__:
                continue
            docstring = module["class"].__doc__
            docstring = "\n    " + docstring.strip()

            doc = " * {0}\n{1}\n".format(name, docstring)
            print(doc)


if __name__ == "__main__":
    ml = ModuleLoader()
    ml.extract_docs()



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class LStrip(Module):
    """Trim whitespace from beginning of current lines."""

    def run(self, app, editor, args):
        # TODO: move cursors in sync with line contents
        line_nums = editor.get_lines_with_cursors()
        for n in line_nums:
            line = editor.lines[n]
            line.data = line.data.lstrip()


module = {
    "class": LStrip,
    "name": "lstrip",
}



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class Lower(Module):
    """Transform current lines to lower case."""

    def run(self, app, editor, args):
        line_nums = []
        for cursor in editor.cursors:
            if cursor.y not in line_nums:
                line_nums.append(cursor.y)
                new_data = editor.lines[cursor.y].get_data().lower()
                editor.lines[cursor.y].data = new_data


module = {
    "class": Lower,
    "name": "lower",
}



#!/usr/bin/env python
import re
from setuptools import setup

version = re.search(
    '^__version__\s*=\s*"([^"]*)"',
    open("suplemon/main.py").read(),
    re.M
    ).group(1)

files = ["config/*.json", "themes/*", "modules/*.py", "linelight/*.py"]

#
# Linage:
#
#      url="https://github.com/bagage/suplemon/"
#
#      url="https://github.com/richrd/suplemon/",
#      author_email="richrd.lewis@gmail.com",
#      author="Richard Lewis",
#

setup(name="Wryten",
      version=version,
      description="Programable Console IDE",
      author="Scott McCallum",
      author_email="sm@intermine.com",
      url="https://github.com/streamcoders/wryten/",
      packages=["suplemon"],
      package_data={"": files},
      include_package_data=True,
      install_requires=[
          "pyperclip",
          "pygments",
          "wcwidth",
          "windows-curses; platform_system=='Windows'"
      ],
      entry_points={
          "console_scripts": ["e=suplemon.cli:main","wryten=suplemon.cli:main"]
      },
      classifiers=[]
      )



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class Upper(Module):
    """Transform current lines to upper case."""

    def run(self, app, editor, args):
        line_nums = []
        for cursor in editor.cursors:
            if cursor.y not in line_nums:
                line_nums.append(cursor.y)
                data = editor.lines[cursor.y].get_data().upper()
                editor.lines[cursor.y].set_data(data)


module = {
    "class": Upper,
    "name": "upper",
}



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class BulkDelete(Module):
    """
    Bulk delete lines and characters.
    Asks what direction to delete in by default.

    Add 'up' to delete lines above highest cursor.
    Add 'down' to delete lines below lowest cursor.
    Add 'left' to delete characters to the left of all cursors.
    Add 'right' to delete characters to the right of all cursors.
    """

    def init(self):
        self.directions = ["up", "down", "left", "right"]

    def handler(self, prompt, event):
        # Get arrow keys from prompt
        if event.key_name in self.directions:
            prompt.set_data(event.key_name)
            prompt.on_ready()
        return True  # Disable normal key handling

    def run(self, app, editor, args):
        direction = args.lower()
        if not direction:
            direction = app.ui.query_filtered("Press arrow key in direction to delete:", handler=self.handler)

        if direction not in self.directions:
            app.set_status("Invalid direction.")
            return False

        # Delete entire lines
        if direction == "up":
            pos = editor.get_first_cursor()
            length = len(editor.lines)
            editor.lines = editor.lines[pos.y:]
            delta = length - len(editor.lines)
            # If lines were removed, move the cursors up the same amount
            if delta:
                editor.move_cursors((0, -delta))

        elif direction == "down":
            pos = editor.get_last_cursor()
            editor.lines = editor.lines[:pos.y+1]

        # Delete from start or end of lines
        else:
            # Select min/max function based on direction
            func = min if direction == "left" else max
            # Get all lines with cursors
            line_indices = editor.get_lines_with_cursors()
            for line_no in line_indices:
                # Get all cursors for the line
                line_cursors = editor.get_cursors_on_line(line_no)
                # Get the leftmost of rightmost x coordinate
                x = func(line_cursors, key=lambda c: c.x).x

                # Delete correct part of the line
                line = editor.lines[line_no]
                if direction == "left":
                    line.data = line.data[x:]
                    # Also move cursors appropriately when deleting left side
                    [c.move_left(x) for c in line_cursors]
                else:
                    line.data = line.data[:x]


module = {
    "class": BulkDelete,
    "name": "bulk_delete",
}



# -*- encoding: utf-8

import time

from suplemon.suplemon_module import Module


class Date(Module):
    """Shows the current date without year in the top status bar."""

    def get_status(self):
        s = time.strftime("%d.%m.")
        if self.app.config["app"]["use_unicode_symbols"]:
            return "" + s
        return s


module = {
    "class": Date,
    "name": "date",
    "status": "top",
}



# -*- encoding: utf-8
"""
Promts based on the Editor class for querying user input.
"""

import os

from .editor import Editor
from .line import Line


# Python 2 compatibility
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


class Prompt(Editor):
    """An input prompt based on the Editor."""
    def __init__(self, app, window):
        Editor.__init__(self, app, window)
        self.ready = False
        self.canceled = False
        self.input_func = lambda: False
        self.caption = ""

    def init(self):
        Editor.init(self)
        # Remove the find feature, otherwise it can be invoked recursively
        del self.operations["find"]

    def set_config(self, config):
        """Set the configuration for the editor."""
        # Override showing line numbers
        config["show_line_nums"] = False
        Editor.set_config(self, config)

    def set_input_source(self, input_func):
        # Set the input function to use while looping for input
        self.input_func = input_func

    def on_ready(self):
        """Accepts the current input."""
        self.ready = True
        return

    def on_cancel(self):
        """Cancels the input prompt."""
        self.set_data("")
        self.ready = True
        self.canceled = True
        return

    def line_offset(self):
        """Get the x coordinate of beginning of line."""
        return len(self.caption)+1

    def render_line_contents(self, line, pos, x_offset, max_len):
        """Render the prompt line."""
        x_offset = self.line_offset()
        # Render the caption
        self.window.addstr(pos[1], 0, self.caption)
        # Render input
        self.render_line_normal(line, pos, x_offset, max_len)

    def handle_input(self, event):
        """Handle special bindings for the prompt."""
        if event.key_name in ["ctrl+c", "escape"]:
            self.on_cancel()
            return False
        if event.key_name == "enter":
            self.on_ready()
            return False

        return Editor.handle_input(self, event)

    def get_input(self, caption="", initial=""):
        """Get text input from the user via the prompt."""
        self.caption = caption
        self.set_data(initial)
        self.end()  # Move to the end of the initial text

        self.refresh()
        success = self.input_loop()
        if success:
            return self.get_data()
        return False

    def input_loop(self):
        # Run the input loop until ready
        while not self.ready:
            event = self.input_func(True)  # blocking
            if event:
                self.handle_input(event)
            self.refresh()
        if not self.canceled:
            return True


class PromptBool(Prompt):
    """An input prompt for booleans based on Prompt."""

    def set_value(self, value):
        self.value = value

    def handle_input(self, event):
        """Handle special bindings for the prompt."""
        name = event.key_name
        if name.lower() == "y":
            self.set_value(True)
            self.on_ready()
            return False
        if name.lower() == "n":
            self.set_value(False)
            self.on_ready()
            return False
        Prompt.handle_input(self, event)

    def get_input(self, caption="", initial=False):
        """Get a boolean value from the user via the prompt."""

        indicator = "[y/N]"
        if initial:
            indicator = "[Y/n]"

        self.caption = caption + " " + indicator
        self.set_value(initial)
        self.end()  # Move to the end of the initial text

        self.refresh()

        # Run the input loop until ready
        success = self.input_loop()
        if success:
            return self.value
        return False


class PromptPassword(Prompt):
    """An input prompt for passwords based on Prompt."""

    def render_line_contents(self, line, pos, x_offset, max_len):
        obscured = Line("*" * len(line))
        Prompt.render_line_contents(self, obscured, pos, x_offset, max_len)


class PromptFiltered(Prompt):
    """An input prompt that allows intercepting and filtering input events."""

    def __init__(self, app, window, handler=None):
        Prompt.__init__(self, app, window)
        self.prompt_handler = handler

    def handle_input(self, event):
        """Handle special bindings for the prompt."""
        # The cancel and accept keys are kept for concistency
        if event.key_name in ["ctrl+c", "escape"]:
            self.on_cancel()
            return False
        if event.key_name == "enter":
            self.on_ready()
            return False

        if self.prompt_handler and self.prompt_handler(self, event):
            # If the prompt handler returns True the default action is skipped
            return True

        return Editor.handle_input(self, event)


class PromptAutocmp(Prompt):
    """An input prompt with basic autocompletion."""

    def __init__(self, app, window, initial_items=[]):
        Prompt.__init__(self, app, window)
        # Whether the autocomplete feature is active
        self.complete_active = False
        # Index of last item that was autocompleted
        self.complete_index = 0
        # Input data to use for autocompletion (stored when autocompletion is activated)
        self.complete_data = ""
        # Default autocompletable items
        self.complete_items = initial_items

    def handle_input(self, event):
        """Handle special bindings for the prompt."""
        name = event.key_name
        if self.complete_active:
            # Allow accepting completed directories with enter
            if name == "enter":
                if self.has_match():
                    self.deactivate_autocomplete()
                    return False
            # Revert autocompletion with esc
            if name == "escape":
                self.revert_autocomplete()
                self.deactivate_autocomplete()
                return False
        if name == "tab":
            # Run autocompletion when tab is pressed
            self.autocomplete()
            # Don't pass the event to the parent class
            return False
        elif name == "shift+tab":
            # Go to previous item when shift+tab is pressed
            self.autocomplete(previous=True)
            # Don't pass the event to the parent class
            return False
        else:
            # If any key other than tab is pressed deactivate the autocompleter
            self.deactivate_autocomplete()
        Prompt.handle_input(self, event)

    def autocomplete(self, previous=False):
        data = self.get_data()  # Use the current input by default
        if self.complete_active:  # If the completer is active use the its initial input value
            data = self.complete_data

        name = self.get_completable_name(data)
        items = self.get_completable_items(data)

        # Filter the items by name if the input path contains a name
        if name:
            items = self.filter_items(items, name)
        if not items:
            # Deactivate completion if there's nothing to complete
            self.deactivate_autocomplete()
            return False

        if not self.complete_active:
            # Initialize the autocompletor
            self.complete_active = True
            self.complete_data = data
            self.complete_index = 0
        else:
            # Increment the selected item countor
            if previous:
                # Go back
                self.complete_index -= 1
                if self.complete_index < 0:
                    self.complete_index = len(items)-1
            else:
                # Go forward
                self.complete_index += 1
                if self.complete_index > len(items)-1:
                    self.complete_index = 0

        item = items[self.complete_index]
        new_data = self.get_full_completion(data, item)
        if len(items) == 1:
            self.deactivate_autocomplete()
        # Set the input data to the completion and move cursor to the end
        self.set_data(new_data)
        self.end()

    def get_completable_name(self, data=""):
        return data

    def get_completable_items(self, data=""):
        return self.complete_items

    def get_full_completion(self, data, item):
        return item

    def has_match(self):
        return False

    def deactivate_autocomplete(self):
        self.complete_active = False
        self.complete_index = 0
        self.complete_data = ""

    def revert_autocomplete(self):
        if self.complete_data:
            self.set_data(self.complete_data)

    def filter_items(self, items, name):
        """Remove items that don't begin with name. Not case sensitive."""
        if not name:
            return items
        name = name.lower()
        return [item for item in items if item.lower().startswith(name)]


class PromptFile(PromptAutocmp):
    """An input prompt with path autocompletion based on PromptAutocmp."""

    def __init__(self, app, window):
        PromptAutocmp.__init__(self, app, window)

    def has_match(self):
        return os.path.isdir(os.path.expanduser(self.get_data()))

    def get_completable_name(self, data=""):
        return os.path.basename(data)

    def get_completable_items(self, data=""):
        return self.get_path_contents(data)  # Get directory listing of input path

    def get_full_completion(self, data, item):
        return os.path.join(os.path.dirname(data), item)

    def get_path_contents(self, path):
        path = os.path.dirname(os.path.expanduser(path))
        # If we get an empty path use the current directory
        if not path:
            try:
                path = os.getcwd()
            except FileNotFoundError:
                # This might happen if the cwd has been
                # removed after starting suplemon.
                return []
        # In case we don't have sufficent permissions
        try:
            contents = os.listdir(path)
        except:
            return []
        contents.sort()
        items = []
        # Append directory separator to directories
        for item in contents:
            if os.path.isdir(os.path.join(path, item)):
                item = item + os.sep
            items.append(item)
        return items



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class Save(Module):
    """Save the current file."""

    def run(self, app, editor, args):
        return app.save_file()


module = {
    "class": Save,
    "name": "save",
}



# -*- encoding: utf-8

import os

from suplemon import config


class SuplemonConfig(config.ConfigModule):
    """Shortcut for openning the config files."""

    def init(self):
        self.config_name = "defaults.json"
        self.config_default_path = os.path.join(self.app.path, "config", self.config_name)
        self.config_user_path = self.app.config.path()


module = {
    "class": SuplemonConfig,
    "name": "config",
}



# -*- encoding: utf-8
"""
Various helper constants and functions.
"""

import os
import re
import sys
import time
import traceback


def curr_time():
    """Current time in %H:%M"""
    return time.strftime("%H:%M")


def curr_time_sec():
    """Current time in %H:%M:%S"""
    return time.strftime("%H:%M:%S")


def multisplit(data, delimiters):
    pattern = "|".join(map(re.escape, delimiters))
    return re.split(pattern, data)


def get_error_info():
    """Return info about last error."""
    msg = "{0}\n{1}".format(str(traceback.format_exc()), str(sys.exc_info()))
    return msg


def get_string_between(start, stop, s):
    """Search string for a substring between two delimeters. False if not found."""
    i1 = s.find(start)
    if i1 == -1:
        return False
    s = s[i1 + len(start):]
    i2 = s.find(stop)
    if i2 == -1:
        return False
    s = s[:i2]
    return s


def whitespace(line):
    """Return index of first non whitespace character on a line."""
    i = 0
    for char in line:
        if char != " ":
            break
        i += 1
    return i


def parse_path(path):
    """Parse a relative path and return full directory and filename as a tuple."""
    if path[:2] == "~" + os.sep:
        p = os.path.expanduser("~")
        path = os.path.join(p+os.sep, path[2:])
    ab = os.path.abspath(path)
    parts = os.path.split(ab)
    return parts


def get_filename_cursor_pos(name):
    default = {
        "name": name,
        "row": 0,
        "col": 0,
    }

    m = re.match(r"(.*?):(\d+):?(\d+)?", name)

    if not m:
        return default

    groups = m.groups()
    if not groups[0]:
        return default

    return {
        "name": groups[0],
        "row": abs(int(groups[1])-1) if groups[1] else 0,
        "col": abs(int(groups[2])-1) if groups[2] else 0,
    }
